/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  MODIF AT&T : Mai 2002 -> en mode supervise, les exemples
    peuvent appartenir a plusieurs classes  */

/*  lia_sct : fonction et type generaux  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <libgram.h>

/*................................................................*/

/* Constantes Generales */

#define TailleLigne	1000

#define False		0
#define True		1

#define NbTypeQuestion	5			/* <?> <+?> <?+> <+?+> + global */
#define NbDirections	2			/* gauche et droite */

#define SUPERVISE	0
#define NON_SUPERVISE	1

#define PPLEX		0
#define GINI		1
#define GINI_LIGHT	2

/*................................................................*/

/*  Fonctions diverses et variees  */

#define min(a,b)        ((a)<(b)?(a):(b))
#define max(a,b)        ((a)>(b)?(a):(b))

void ERREUR(char *,char *);

void *my_malloc(int);

/*................................................................*/

/*  Structure de codage des noeuds du CART  */

/* MODIF 0500 :

Les mots sont decrit a partir de plusieurs niveaux de representation
decrits dans un fichier de configuration. Ainsi, ty_node code
la question ainsi que son niveau dans les differentes representations
choisies

*/

typedef struct type_node
	{
	char	t_noeud,	/* Feuille ou noeud interne */
		t_question,	/* Type de question pose : J,U,R,L */
		qloc,		/* Si cette question se pose a droite ou a gauche du KS */
		KS,		/* Le type du noeud d'ou l'on vient */
		niveau_q;	/* Niveau de la question ds la representation des mots */
	wrd_index_t	q;	/* La question pose : cad le mot lui meme */
	double *proba;		/* Repartition des probas selon les etiquettes (mode SUPERVISE) */
	struct type_node *fg,*fd; /* fg=oui fd=non */
    int cate;		/* Num du noeud, en absolu, par ordre de creation */
	} *ty_node;


/* stockage des infos sur le type de format de representation choisi */

typedef struct type_format
	{
	int nb_niveaux;			/* nb de niveaux de representation */
	int nb_questall;		/* nb of questions related to the whole string */
	ty_lexique *tablex;		/* tableau des lexiques */
	ty_lexique lex_reference;	/* lexique des mots references */
	ty_lexique lex_etiquette;	/* lexique des etiquettes possibles des exemples */
	int indice_ML;			/* indice du niveau ou le ML sera estime (si non supervise) */
	char mode;			/* SUPERVISE / NON_SUPERVISE */
	char critere;			/* PPLEX / GINI / .... */
	int nb_etiquette;		/* nb d'etiquettes en mode SUPERVISE */
	double delta;			/* minimum de gain sur le critere entre pere et fils */
	} *ty_format;

/*  Codage d'un item sous la forme d'une liste doublement chainee
    d'unitees representant des niveaux differents : phon, cohortes, etc.  */

typedef struct type_item
	{
	char inKS;
	wrd_index_t *code; /* tableaux de codes suivant les diff niveaux de representation */
	struct type_item *suiv,*pred;
	} *ty_item;

typedef struct
	{
	char gap_avant,gap_apres;
	ty_item ptKS;
	} type_KS;

typedef struct type_sample
	{
	int numero;		/* numero de l'exemple (incremental) */
	type_KS KS;		/* KS en cours sur l'exemple */
	ty_item item;		/* liste chainee codant l'exemple */
	wrd_index_t ref; 	/* indice de l'identificateur de l'exemple */
	double nb_occu;		/* nb d'occurence de l'exemple dans le corpus */
	wrd_index_t *classe;	/* indice des classes representant l'exemple */
	wrd_index_t *questall;	/* table containing the info related to the whole string */
	struct type_sample *suiv;
	} *ty_sample;


/*  Codage des items ayant repondu oui a la question  */
/*  Cela permet d'eviter de poser a nouveau les questions  */

typedef struct
	{
	char good;
	ty_item ptKS;
	} type_succes;

/*  Codage de la liste de questions  */

typedef struct
	{
	wrd_index_t q;
	char t_question,qloc,gap_avant,gap_apres,niveau_q;
	type_succes *bilan,*bilan_devt;
	int nb,nb_devt;
	double critere_oui,critere_pere_oui,critere_non,critere_pere_non;
	} type_question;

/*................................................................*/

/* Variables Globales */

extern int Debug;				/* niveau de Debug */
extern int MaxHauteur;				/* hauteur maxi de l'arbre construit ou lu */
extern ty_format MyFormat;			/* descripteur du format des donnees traitees */
extern char TablTypeQuestion[NbTypeQuestion];	/* code des questions : J L R U */
extern char TablDirection[NbDirections];	/* code des directions : r l */
extern type_question *Question;			/* tableau de questions */
extern int NbQuestion;				/* nb de questions stockees dans le tableau */
extern double MaxCritere;                       /* quantite maximale pouvant recevoir le critere choisi */
extern int NUMBER_NODE;		/* Node ID */

/*................................................................*/

/*  Fonctions d'affichage  */

/* affichage de l'arbre en prenant une racine et un fichier */
void AfficheArbre(ty_node, FILE *, ty_format);

/* affichage d'une question */
void AfficheQuestion(type_question, ty_format);

/* affichage d'une liste d'exexemple d'apprentissage */
void AfficheListSample(ty_sample, ty_format);

/* affichage d'un item */
void affiche_item(ty_item, ty_format);

/*................................................................*/

/*  Allocation memoire  */

ty_node nouveau_node(char,char,int,char,char,wrd_index_t,char,double *);
ty_item nouveau_item(wrd_index_t*,char,ty_item,ty_format);
void libere_item(ty_item);
void libere_list_item(ty_item);
ty_sample nouveau_sample(wrd_index_t,int,type_KS,ty_item,double,wrd_index_t*,wrd_index_t*,ty_sample);
void libere_format(ty_format);

/*................................................................*/

/* renvoi une chaine de caractere corrspondant a un code et un niveau de representation */
char *renvoi_chaine(wrd_index_t, int, ty_format);

/* renvoi une chaine de caractere corrspondant a un code et un lexique */
char *renvoi_chaine_lex(wrd_index_t, ty_lexique );

/*................................................................*/

/* lecture du fichier de description des formats de fichier .arbre.desc_lex */

ty_format LectureDescripteurFormat(char *);

/*................................................................*/


/* code un item a partir d'une chaine representant un exemple */
ty_item CodeItem(char *, ty_format);
 
