/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  lia_sct_build : gestion de la construction des arbres  */

/*  MODIF March 26 2002 : Now we compute the reduction in perplexity
    between ML YES and NO and the ML of the father on the SAME
    corpus (either YES or NO) !!!!  */

/*  MODIF August 2002 : adding global questions to the CART  */

/*  MODIF October 2004 : now, nb_occu is a double (for Boosting)
 *    and the field 'cate' in ty_node is the ID of the node  */

#include <lia_sct.h>
#include <lia_sct_pplex.h>

/*................................................................*/

/* codage des variables necessaires au critere GINI */

double *RepartEtiquetteOUI=NULL,*RepartEtiquetteNON=NULL;

double calcule_gini(int taille,double *repart,ty_format my_format)
{
double impt;
int n,k;

if (taille==0) return (double)0;

for(n=0,impt=(double)0;n<my_format->nb_etiquette;n++)
 for(k=0;k<my_format->nb_etiquette;k++)
  if (n!=k) impt+=(repart[n]/(double)taille)*(repart[k]/(double)taille);
return impt;
}

double calcule_gini_corpus(ty_sample corpus,int taille,ty_format my_format)
{
int i;
ty_sample pt;

for(i=0;i<my_format->nb_etiquette;i++) RepartEtiquetteOUI[i]=(double)0;
for(pt=corpus;pt;pt=pt->suiv)
 for (i=0;i<my_format->nb_etiquette;i++)
  if (pt->classe[i]>0) RepartEtiquetteOUI[pt->classe[i]-1]+=pt->nb_occu;
return calcule_gini(taille,RepartEtiquetteOUI,my_format);
}

double *calcule_repart_proba(ty_sample corpus,int taille,ty_format my_format)
{
double *proba,somm=0;
int i;
ty_sample pt;

proba=(double *)malloc(sizeof(double)*my_format->nb_etiquette);
for(i=0;i<my_format->nb_etiquette;i++) proba[i]=(double)0;
if (taille>0)
 {
 for(pt=corpus;pt;pt=pt->suiv)
  {
  for (i=0;i<my_format->nb_etiquette;i++)
   if (pt->classe[i]>0) proba[pt->classe[i]-1]+=pt->nb_occu;
  somm+=pt->nb_occu;
  }
 for(i=0;i<my_format->nb_etiquette;i++) proba[i]/=somm;
 }
return proba;
}

/*................................................................*/

/*  Codage des exemples d'apprentissage  */

/* codage dans un tableau de sample */
ty_sample *CorpusTraining,*CorpusDevt,*CorpusTrainingLocal,*CorpusDevtLocal;
int NbSampleTraining=0,NbSampleDevt=0;

/*................................................................*/

/* Gestion des questions a poser pour construire l'arbre */

type_question *Question;                 /* tableau de questions */
int NbQuestion;                          /* nb de questions stockees dans le tableau */

void AfficheQuestion(type_question quest,ty_format my_format)
{
fprintf(stdout,"Question : %c %d %s (%d) %d sur la branche %c\n",quest.t_question,
  quest.gap_avant,renvoi_chaine(quest.q,quest.t_question=='G'?my_format->nb_niveaux:quest.niveau_q,my_format),
  quest.niveau_q,quest.gap_apres,quest.qloc);
}

/*................................................................*/

/*  Chargement des ressources corpus et liste des questions  */

/* le nouveau format des questions est le suivant :
<graf>\t<niveau question>  : avec les niveaux commencant a 1
ATTENTION : pas de lignes vides !!
*/

void LectureListQuestion(char *chnom,ty_format my_format)
{
FILE *file;
char ch[TailleLigne],*quest,*chniv;
wrd_index_t *list_code,*list_glb_code;
int *list_niveaux;
int n,i,j,k,nb_quest,nb_glb_quest;

sprintf(ch,"%s.glb_quest",chnom);
if (file=fopen(ch,"r"))
 {
 for (nb_glb_quest=0;fgets(ch,TailleLigne,file);nb_glb_quest++);
 fseek(file,0,SEEK_SET);
 list_glb_code=(wrd_index_t *)my_malloc(sizeof(wrd_index_t)*nb_glb_quest);
 for(nb_glb_quest=0;fgets(ch,TailleLigne,file);nb_glb_quest++)
  {
  strtok(ch," \t\n");
  quest=(char*)strtok(NULL," \t\n");
  if (!Mot2Code(quest,&(list_glb_code[nb_glb_quest]),my_format->tablex[my_format->nb_niveaux]))
   ERREUR("Mauvais code global pour la question :",quest);
  }
 fclose(file);
 }
else
 {
 fprintf(stderr,"\t-> no global questions\n");
 nb_glb_quest=0;
 list_glb_code=NULL;
 }

sprintf(ch,"%s.question",chnom);
if (!(file=fopen(ch,"r"))) ERREUR("je ne peux ouvrir ",ch);
for (nb_quest=0;fgets(ch,TailleLigne,file);nb_quest++);
fseek(file,0,SEEK_SET);
list_code=(wrd_index_t *)my_malloc(sizeof(wrd_index_t)*nb_quest);
list_niveaux=(int *)my_malloc(sizeof(int)*nb_quest);
for(nb_quest=0;fgets(ch,TailleLigne,file);nb_quest++)
 {
 quest=(char*)strtok(ch," \t\n");
 chniv=(char*)strtok(NULL," \t\n");
 if (!strcmp(quest,"XX"))
  {
  fprintf(stderr,"WARNING: the symbol 'XX' has been found in the list of questions, it will be discarded");
  nb_quest--;
  }
 else
  {
  sscanf(chniv,"%d",&(list_niveaux[nb_quest]));
  list_niveaux[nb_quest]--; /* on commence a 0 maintenant */
  if (!Mot2Code(quest,&(list_code[nb_quest]),my_format->tablex[list_niveaux[nb_quest]]))
   ERREUR("Mauvais code pour la question :",quest);
  }
 }
fclose(file);

NbQuestion=nb_glb_quest+nb_quest*(NbTypeQuestion-1)*NbDirections; /* Pour les questions : J L R U en gauche et droite */
if (Debug) fprintf(stderr,"\t\t- Lecture d'un lexique de %d mots\n",nb_quest);
if (Debug) fprintf(stderr,"\t\t- Construction d'un ensemble de %d questions\n",NbQuestion);

/*  Initialisation du tableau de questions  */
Question=(type_question*)my_malloc(sizeof(type_question)*NbQuestion);
for(k=n=0;n<nb_quest;n++)
 for(i=0;i<NbTypeQuestion-1;i++)
  for(j=0;j<NbDirections;j++)
   {
   Question[n*8+i*2+j].q=list_code[n];
   Question[n*8+i*2+j].niveau_q=list_niveaux[n];
   Question[n*8+i*2+j].t_question=TablTypeQuestion[i];
   switch (TablTypeQuestion[i])
    {
    case 'J' : Question[n*8+i*2+j].gap_avant=0; Question[n*8+i*2+j].gap_apres=0; break;
    case 'L' : Question[n*8+i*2+j].gap_avant=0; Question[n*8+i*2+j].gap_apres=1; break;
    case 'R' : Question[n*8+i*2+j].gap_avant=1; Question[n*8+i*2+j].gap_apres=0; break;
    case 'U' : Question[n*8+i*2+j].gap_avant=1; Question[n*8+i*2+j].gap_apres=1; break;
    default  : ERREUR("C'est quoi cette question ????","");
    }
   Question[n*8+i*2+j].qloc=TablDirection[j];
   Question[n*8+i*2+j].bilan=Question[n*8+i*2+j].bilan_devt=NULL;
   Question[n*8+i*2+j].nb=Question[n*8+i*2+j].nb_devt=0;
   Question[n*8+i*2+j].critere_oui=Question[n*8+i*2+j].critere_pere_oui=MaxCritere;
   Question[n*8+i*2+j].critere_non=Question[n*8+i*2+j].critere_pere_non=MaxCritere;
   k++;
   }
for(n=0;n<nb_glb_quest;n++)
 {
 Question[n+k].q=list_glb_code[n];
 Question[n+k].niveau_q=0;
 Question[n+k].t_question='G'; /* question global */
 Question[n+k].gap_avant=Question[n+k].gap_apres=0;	/* undefined */
 Question[n+k].qloc=TablDirection[0];			/* undefined */
 Question[n+k].bilan=Question[n+k].bilan_devt=NULL;
 Question[n+k].nb=Question[n+k].nb_devt=0;
 Question[n+k].critere_oui=Question[n+k].critere_pere_oui=MaxCritere;
 Question[n+k].critere_non=Question[n+k].critere_pere_non=MaxCritere;
 }
if ((n+k)!=NbQuestion) { fprintf(stderr,"ERROR: NbQuestion mismatch (%d instead of %d)\n",n+k,NbQuestion); exit(0); }
free(list_code);
free(list_niveaux);
if (list_glb_code) free(list_glb_code);
}

/*................................................................*/

/* Lecture du corpus et construction des structures des ngrams */
/* MODIF : lecture de fichiers au format :

i=item (l items), n=niveau (k niveaux)

<graf> <i1 n1> <i1 n2> .... <i1 nk> <i2 n1> <i2 n2> .... <i2 nk> .... <ilnk>

*/

ty_sample *lecture_corpus(char *chnom,int *nb_sample,ty_format my_format)
{
FILE *file;
char ch[10*TailleLigne],*graf,*chitem,*pafini,*chid;
type_KS KS;
ty_sample *corpus;
int n,i;
double nb_occu;
wrd_index_t i_graf,*classe,iclas,*questall;
ty_item item;

KS.gap_avant=0; KS.gap_apres=1; KS.ptKS=NULL;
if (!(file=fopen(chnom,"r"))) ERREUR("je ne peux ouvrir ",chnom);
for ((*nb_sample)=0;fgets(ch,10*TailleLigne,file);) (*nb_sample)++;
fseek(file,0,SEEK_SET);
i=1;
if (my_format->mode==SUPERVISE) i=2; /* on /2 a cause des lignes etiquettes */
if (my_format->nb_questall>0)   i++; /* on /3 ou /2 because global questions */
(*nb_sample)/=i;
corpus=(ty_sample*)my_malloc(sizeof(ty_sample)*(*nb_sample));

for(n=0,pafini=fgets(ch,10*TailleLigne,file);(n<(*nb_sample))&&(pafini);)
 {
 chid=(char*)strtok(ch," \t\n");
 /*fprintf(stderr,"chid=%s\n",chid);*/

 for(i=strlen(chid)+1;(chid[i]==' ')||(chid[i]=='\t');i++); chitem=chid+i;
 if (my_format->lex_reference==NULL) ERREUR("No lexique reference !!","");

 if (!Mot2Code(chid,&i_graf,my_format->lex_reference))
  ERREUR("Mauvais code pour le mot :",chid);
 item=CodeItem(chitem,my_format);

 pafini=fgets(ch,10*TailleLigne,file);

 if ((pafini)&&(ch[0]=='&')) /* global questions */
  {
  questall=(wrd_index_t*)malloc(sizeof(wrd_index_t)*my_format->nb_questall);
  for(i=0,graf=strtok(ch+2," \t\n");graf;graf=strtok(NULL," \t\n"),i++)
   {
   if (i==my_format->nb_questall) ERREUR("bad format in :",chid);
   if (Mot2Code(graf,&iclas,my_format->tablex[my_format->nb_niveaux]))
    questall[i]=iclas;
   else
    { fprintf(stderr,"ERREUR: this graf should be in QUEST_GLB lexicon number %d : %s\n",my_format->nb_niveaux+1,graf); exit(0); }
   }
  for(;i<my_format->nb_questall;i++) questall[i]=0;
  pafini=fgets(ch,10*TailleLigne,file);
  }

 if (my_format->mode==SUPERVISE)
  {
  /* on lit l'etiquette et le nombre d'occurence */
  if ((pafini==NULL)||(ch[0]!='@')) ERREUR("il manque une etiquette !!","");
  classe=(wrd_index_t*)malloc(sizeof(wrd_index_t)*my_format->nb_etiquette);
  for(i=0;i<my_format->nb_etiquette;i++) classe[i]=0; /* the indice of label lexicon start at 1 */
  for(i=0,graf=strtok(ch+2," \t\n");graf;graf=strtok(NULL," \t\n"),i++)
   {
   if (Mot2Code(graf,&iclas,my_format->lex_etiquette))
    {
    if (i>=my_format->nb_etiquette) ERREUR("why more label for this sample than in the desc file ? ->",graf);
    classe[i]=iclas;
    }
   else if (sscanf(graf,"%lf",&nb_occu)!=1) ERREUR("Bad field in the label line:",graf);
   }
  pafini=fgets(ch,10*TailleLigne,file);
  }
 else { nb_occu=0; classe=NULL; }
 corpus[n++]=nouveau_sample(i_graf,n,KS,item,nb_occu,classe,questall,NULL);
 }
fclose(file);

/* on chaine les cases du tableau en une liste d'exemples */
for(n=0;n<(*nb_sample);n++)
 if (n<(*nb_sample)-1) corpus[n]->suiv=corpus[n+1];
 else corpus[n]->suiv=NULL;
return corpus;
}

void lecture_corpus_training(char *nomfich,ty_format my_format)
{
CorpusTraining=lecture_corpus(nomfich,&NbSampleTraining,my_format);
CorpusTrainingLocal=(ty_sample*)my_malloc(sizeof(ty_sample)*NbSampleTraining);

if (my_format->mode==NON_SUPERVISE)
 {
 RacineOuiGram=rempli_niveaux(CorpusTraining[0],my_format);
 RacineNonGram=rempli_niveaux(CorpusTraining[0],my_format);
 RacineAllGram=rempli_niveaux(CorpusTraining[0],my_format);
 if (Debug>=2)
  fprintf(stderr,"\t\tperplexite du corpus de depart (sur l'apprentissage) = %.2lf\n",calcule_perplexity(RacineOuiGram));
 }
}

void lecture_corpus_devt(char *nomfich,ty_format my_format)
{
if (!strcmp(nomfich,"NULL")) CorpusDevt=CorpusDevtLocal=NULL;
else
 {
 CorpusDevt=lecture_corpus(nomfich,&NbSampleDevt,my_format);
 CorpusDevtLocal=(ty_sample*)my_malloc(sizeof(ty_sample)*NbSampleDevt);
 if (my_format->mode==NON_SUPERVISE)
  if (Debug>=2) fprintf(stderr,"\t\tperplexite du corpus de test de depart (sur le devt) = %.2lf\n",
	calcule_perplexity_test(RacineOuiGram,CorpusDevt[0],0,my_format));
 }
}

/*................................................................*/

/*  Creation de l'arbre en direct live  */

ty_item PoseQuestion(ty_sample sample,int indice_question, ty_format my_format)
{
ty_item pt,tmp;
type_KS KS;
wrd_index_t question;
char niveau_q, qloc, gap_avant, gap_apres;

/* init */
KS=sample->KS;
question=Question[indice_question].q;
niveau_q=Question[indice_question].niveau_q;
qloc=Question[indice_question].qloc;
gap_avant=Question[indice_question].gap_avant;
gap_apres=Question[indice_question].gap_apres;

/* first, is it a global question */
if (Question[indice_question].t_question=='G')
 {
 int i;
 for(i=0;(i<my_format->nb_questall)&&(sample->questall[i]!=question);i++);
 if (i<my_format->nb_questall) /* OK, we've got a match */
  return sample->item;
 else
  return NULL;
 }

if (qloc=='l')  /* on cherche a la gauche du KS */
 {   /* sur la racine, ptKS est egal a NULL */

 /* test de coherence */
 /* si le KS.gap_avant==0, on change de KS en reculant vers la gauche, jusqu'a trouver un gap */
 if (KS.gap_avant==0)
  {
  for(pt=KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->pred);
  if (pt==NULL) return NULL; /* Il n'y a pas de GAP sur la gauche !!!! */
  }
 else
  if (KS.ptKS==NULL) pt=sample->item; else pt=KS.ptKS->pred;

 if (pt==NULL) return NULL;

 if (gap_apres==0)
  { /* la question doit se trouver juste avant le KS */
  if ((pt->inKS==0)&&(pt->code[niveau_q]==question))
   {
   /* on test s'il ne doit pas y avoir de GAP sur gauche */
   if ((gap_avant==0)&&(pt->pred)&&(pt->pred->inKS==0)) return NULL;
   }
  else return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->pred;
  while ((pt)&&(pt->inKS==0)&&(pt->code[niveau_q]!=question)) pt=pt->pred;

  if ((pt==NULL)||(pt->inKS!=0)) return NULL;
  /* on test s'il ne doit pas y avoir de GAP sur gauche */
  if ((gap_avant==0)&&(pt->pred)&&(pt->pred->inKS==0)) return NULL;
  }

 if (gap_avant==0) /* il ne faut pas qu'il y est de trou avant */
  {
  for(tmp=pt->pred;(tmp)&&(tmp->inKS==1);tmp=tmp->pred);
  if (tmp) return NULL;
  }
 return pt;
 }
else
 {  /* on cherche a la droite du KS */

 /* test de coherence */
 /* si le KS.gap_apres==0, on change de KS en avancant vers la droite, jusqu'a trouver un gap */
 if (KS.gap_apres==0)
  {
  for(pt=KS.ptKS;(pt)&&(pt->inKS==1);pt=pt->suiv);
  if (pt==NULL) return NULL; /* Il n'y a pas de GAP sur la droite !!!! */
  }
 else if (KS.ptKS==NULL) pt=sample->item; else pt=KS.ptKS->suiv;

 if (pt==NULL) return NULL;

 if (gap_avant==0)
  { /* la question doit se trouver juste apres le KS */
  if ((pt->inKS==0)&&(pt->code[niveau_q]==question))
   {
   /* on test s'il ne doit pas y avoir de GAP sur droite */
   if ((gap_apres==0)&&(pt->suiv)&&(pt->suiv->inKS==0)) return NULL;
   }
  else return NULL;
  }
 else
  { /* il peut y avoir un gap entre la question et le KS */
  /* MODIF : il DOIT y avoir un gap */
  pt=pt->suiv;
  while ((pt)&&(pt->inKS==0)&&(pt->code[niveau_q]!=question)) pt=pt->suiv;
  if ((pt==NULL)||(pt->inKS!=0)) return NULL;
  /* on test s'il ne doit pas y avoir de GAP sur droite */
  if ((gap_apres==0)&&(pt->suiv)&&(pt->suiv->inKS==0)) return NULL;
  }

 if (gap_apres==0) /* il ne faut pas qu'il y est de trou apres */
  {
  for(tmp=pt->suiv;(tmp)&&(tmp->inKS==1);tmp=tmp->suiv);
  if (tmp) return NULL;
  }
 return pt;
 }
}

void calcule_critere_pplex(int n,int taille_corpus,int taille_corpus_devt,ty_format my_format)
{
long nb_ouigram,nb_nongram,nb_allgram;
int i;

/* on remet a zero les ngram */
raz_gram(RacineOuiGram);
raz_gram(RacineNonGram);
raz_gram(RacineAllGram);
nb_ouigram=nb_nongram=nb_allgram=0;

for (i=0;i<taille_corpus;i++)
 {
 if (Question[n].bilan[i].good==True)
  RacineOuiGram=ajoute_item_ngram(CorpusTrainingLocal[i]->item,RacineOuiGram,&nb_ouigram,my_format);
 else
  RacineNonGram=ajoute_item_ngram(CorpusTrainingLocal[i]->item,RacineNonGram,&nb_nongram,my_format);
 RacineAllGram=ajoute_item_ngram(CorpusTrainingLocal[i]->item,RacineAllGram,&nb_allgram,my_format);
 }

if (Debug>=2) printf("on remplo les probas OUI\n");
rempli_proba(nb_ouigram,RacineOuiGram,0);
if (Debug>=2) printf("on remplo les probas NON\n");
rempli_proba(nb_nongram,RacineNonGram,0);
if (Debug>=2) printf("on remplo les probas ALL\n");
rempli_proba(nb_allgram,RacineAllGram,0);

if (CorpusDevt==NULL) /* pp sur l'apprentissage */
 {
 Question[n].critere_oui=calcule_perplexity(RacineOuiGram);
 Question[n].critere_pere_oui=calcule_perplexity_test_succes(True,RacineAllGram,CorpusTrainingLocal,
							Question[n].bilan,taille_corpus,Debug>2?1:0,my_format);
 }
else
 {
 if (Debug>=2) printf("on calcule la perplexity OUI sur le devt\n");
 if (Question[n].nb_devt>0)
  {
  Question[n].critere_oui=calcule_perplexity_test_succes(True,RacineOuiGram,CorpusDevtLocal,
						   	 Question[n].bilan_devt,
						         taille_corpus_devt,Debug>2?1:0,my_format);
  Question[n].critere_pere_oui=calcule_perplexity_test_succes(True,RacineAllGram,CorpusDevtLocal,
							 Question[n].bilan_devt,
							 taille_corpus_devt,Debug>2?1:0,my_format);
  if (Debug>=2)
   printf("\toui -> PP_oui(devt_oui)=%.2lf  PP_pere(devt_oui)=%.2lf  gain=%.2lf\n",
    Question[n].critere_oui,Question[n].critere_pere_oui,Question[n].critere_pere_oui-Question[n].critere_oui);
  }
 else
  {
  if (Debug>=2) printf("liste d'exemple devt succes = NULL\n");
  Question[n].critere_oui=Question[n].critere_pere_oui=MaxCritere;
  }
 }

if (CorpusDevt==NULL) /* pp sur l'apprentissage */
 {
 Question[n].critere_non=calcule_perplexity(RacineNonGram);
 Question[n].critere_pere_non=calcule_perplexity_test_succes(False,RacineAllGram,CorpusTrainingLocal,
							Question[n].bilan,taille_corpus,Debug>2?1:0,my_format);
 }
else
 {
 if (Debug>=2) printf("on calcule la perplexity NON sur le devt\n");
 if (Question[n].nb_devt<taille_corpus_devt)
  {
  Question[n].critere_non=calcule_perplexity_test_succes(False,RacineNonGram,CorpusDevtLocal,
							 Question[n].bilan_devt,
							 taille_corpus_devt,Debug>2?1:0,my_format);
  Question[n].critere_pere_non=calcule_perplexity_test_succes(False,RacineAllGram,CorpusDevtLocal,
							 Question[n].bilan_devt,
							 taille_corpus_devt,Debug>2?1:0,my_format);
  if (Debug>=2)
   printf("\tnon -> PP_non(devt_non)=%.2lf  PP_pere(devt_non)=%.2lf  gain=%.2lf\n",
    Question[n].critere_non,Question[n].critere_pere_non,Question[n].critere_pere_non-Question[n].critere_non);
  }
 else
  {
  if (Debug>=2) printf("liste d'exemple devt echec = NULL\n");
  Question[n].critere_non=Question[n].critere_pere_non=MaxCritere;
  }
 }
}

void calcule_critere_gini(int n,int taille_corpus,int taille_corpus_devt,ty_format my_format)
{
int i,j,taille,nb;
ty_sample *corpus;
type_succes *bilan;

for(i=0;i<my_format->nb_etiquette;i++) RepartEtiquetteOUI[i]=RepartEtiquetteNON[i]=(double)0;

if (CorpusDevt==NULL)
 {
 corpus=CorpusTrainingLocal; taille=taille_corpus; nb=Question[n].nb; bilan=Question[n].bilan;
 }
else
 {
 corpus=CorpusDevtLocal; taille=taille_corpus_devt; nb=Question[n].nb_devt; bilan=Question[n].bilan_devt;
 }

/*printf("GINI: %s taille=%d\n",CorpusDevt==NULL?"TRAIN":"DEV",taille);*/

for (i=0;i<taille;i++)
 {
 if (bilan[i].good==True)
  {
  /*printf("\t sample OK:");*/
  for(j=0;j<my_format->nb_etiquette;j++)
   if (corpus[i]->classe[j]>0)
    {
    /*printf(" CLASSE(%d)",corpus[i]->classe[j]);*/
    RepartEtiquetteOUI[corpus[i]->classe[j]-1]+=corpus[i]->nb_occu;
    }
  /*printf("\n");*/
  }
 else
  {
  for(j=0;j<my_format->nb_etiquette;j++)
   if (corpus[i]->classe[j]>0) RepartEtiquetteNON[corpus[i]->classe[j]-1]+=corpus[i]->nb_occu;
  }
 }
Question[n].critere_oui=nb>0?calcule_gini(nb,RepartEtiquetteOUI,my_format):MaxCritere;
Question[n].critere_non=nb<taille?calcule_gini(taille-nb,RepartEtiquetteNON,my_format):MaxCritere;
}

/* calcule le critere (devt ou app) sur la question n */
void calcule_critere(int n,int taille_corpus,int taille_corpus_devt,ty_format my_format)
{
if (my_format->mode==NON_SUPERVISE)
 calcule_critere_pplex(n,taille_corpus,taille_corpus_devt,my_format);
else
 calcule_critere_gini(n,taille_corpus,taille_corpus_devt,my_format);
}

int choisit_question_pplex(int taille_corpus,int taille_corpus_devt,double *best,ty_format my_format)
{
int taille_inf,taille_sup,ntest,nchoix,n;
double eval,gain1,gain2;

fprintf(stderr,"XXXXXXXX : CORRIGER TAILLE_CORPUS == TAILLE_CORPUS_DEVT SI DEV !!!!\n\n");
exit(0);

/* on veut couper 'presque' en 2 le corpus */

taille_inf=(int)((double)taille_corpus/(double)2-((double)taille_corpus/(double)2)*(double)0.2);
taille_sup=(int)((double)taille_corpus/(double)2+((double)taille_corpus/(double)2)*(double)0.2);

/* on prends la question qui produit le + gros gain en  perplexite entre le ML_oui, le ML_non et le LM_oui+non
   sur les 2 corpus C_oui et C_non */

for(gain1=gain2=0,ntest=nchoix=-1,n=0;n<NbQuestion;n++)
 {
 eval=(Question[n].critere_pere_oui-Question[n].critere_oui)+(Question[n].critere_pere_non-Question[n].critere_non);
 if (
     (eval>=gain1) 											&&
     (Question[n].critere_oui<=(Question[n].critere_pere_oui+(Question[n].critere_pere_oui/4)))	&&
     (Question[n].critere_non<=(Question[n].critere_pere_non+(Question[n].critere_pere_non/4)))	&&
     (Question[n].nb>=taille_inf)&&(Question[n].nb<=taille_sup)
    )
  { gain1=eval; nchoix=n; }
 
  if (
     (eval>=gain2)											&&
     (Question[n].critere_oui<=(Question[n].critere_pere_oui+(Question[n].critere_pere_oui/10)))	&&
     (Question[n].critere_non<=(Question[n].critere_pere_non+(Question[n].critere_pere_non/10)))
     )
   { ntest=n; gain2=eval; }
 }
if (Debug>=2) printf("choisit_question : nchoix=%d ntest=%d gain1=%.2lf gain2=%.2lf\n",nchoix,ntest,gain1,gain2);

/* si on ne trouve pas de question qui coupe au mieux en deux, on prends simplement la meilleure */
/*if (nchoix<0)*/ { nchoix=ntest; gain1=gain2; }

if ((nchoix!=-1)&&(gain1<=my_format->delta)) *best=MaxCritere;

*best=(Question[nchoix].critere_oui+Question[nchoix].critere_non)/(double)2;

return nchoix;
}

int choisit_question_gini(double critere_pere,int taille_corpus,int taille_corpus_devt,
		                  double *best,ty_format my_format)
{
double delta;
int n,nchoix=0,nb,taille;

/* on prends la question qui provoque le plus gros gain en perte d'impurite */
*best=-MaxCritere;

if (Debug>=2) printf("Dans choisit_question_gini: taille_corpus=%d\n",taille_corpus);

for(n=0;n<NbQuestion;n++)
 {
 if (CorpusDevt==NULL) { nb=Question[n].nb; taille=taille_corpus; }
 else { nb=Question[n].nb_devt; taille=taille_corpus_devt; }

 if ((nb>0)&&((taille-nb)>0))
  {
  if (my_format->critere==GINI)
   delta=critere_pere-(Question[n].critere_oui*(double)nb)/(double)taille
		   -(Question[n].critere_non*(double)(taille-nb))/(double)taille;
  else
  if (my_format->critere==GINI_LIGHT)
   delta=critere_pere-(Question[n].critere_oui*0.5)-(Question[n].critere_non*0.5);
  else ERREUR("Unknown criterion !!","");
  if (Debug>=2)
   printf("\t Question[%d].nb=%d - Question[%d].critere_oui=%.2lf - Question[%d].critere_non=%.2lf - DELTA=%.2lf\n",
		 n,nb,n,Question[n].critere_oui,n,Question[n].critere_non,delta);
  }
 else delta=-MaxCritere;
 if (delta>*best) { *best=delta; nchoix=n; }
 }

if (Debug>=2)
 printf("\t choisit_question_gini=%d (best=%.2lf,min delta=%.2lf) avec\
 critere_pere=%.2lf critere_oui=%.2lf critere_non=%.2lf nb=%d taille_corpus=%d\n",
	nchoix,*best,my_format->delta,critere_pere,Question[nchoix].critere_oui,
	Question[nchoix].critere_non,CorpusDevt==NULL?Question[nchoix].nb:Question[nchoix].nb_devt,
	CorpusDevt==NULL?taille_corpus:taille_corpus_devt);

if ((*best<=my_format->delta)||(*best==-MaxCritere)) *best=MaxCritere;

return nchoix;
}

int choisit_question(double critere_pere,int taille_corpus,int taille_corpus_devt,
		             double *best,ty_format my_format)
{
if (my_format->mode==NON_SUPERVISE)
 return choisit_question_pplex(taille_corpus,taille_corpus_devt,best,my_format);
else
 return choisit_question_gini(critere_pere,taille_corpus,taille_corpus_devt,best,my_format);
}

ty_node CreationTree(
  		    char t_question_prec,	/* type de la question precedente (pere) */
		    ty_sample corpus,		/* corpus d'apprentissage */
		    int taille_corpus,		/* taille du corpus a l'entree ds la fonction */
		    ty_sample corpus_devt,	/* corpus de developpement, ou NULL */
		    int taille_corpus_devt,	/* taille du corpus developpement a l'entree ds la fonction */
		    int hauteur,		/* hauteur de l'arbre en cours */
		    int nb_elem_feuille,	/* nb mini d'elt par feuille */
		    int haut_maxi,		/* hauteur maxi de l'arbre */
		    double critere_pere,	/* quantite (pplex, impurite) liee au noeud pere */
		    ty_format my_format		/* descripteur du SCT (lex, etc) */
		    )
{
int n,nb_corpus,nchoix,nb_oui,nb_non,nb_oui_devt,nb_non_devt,i;
ty_sample pt,good,baad,devt_good,devt_baad,ptgood,ptbaad;
ty_item resu;
ty_node racine;
double best,critere_oui,critere_non;
type_succes *bilan_tmp;

MaxHauteur=max(MaxHauteur,hauteur);

if (Debug>=2)
 printf("GROW_TREE : Arbre de %d niveaux - Corpus reduit a %d elements - Prof maxi=%d - critere=%.2lf\n",
	hauteur,taille_corpus,MaxHauteur,critere_pere);

/* Si le corpus contient nb_elem_feuille samples ou moins, on fait une feuille */
/* Si la prof de l'arbre est egale a haut_maxi, on fait une feuille */
if ((taille_corpus<=2*nb_elem_feuille)||(hauteur>=haut_maxi))
 {
 return nouveau_node(t_question_prec,'L',NUMBER_NODE++,'0','0',0,'0', /* 0 correspond au mot inconnu */
	my_format->mode==SUPERVISE?calcule_repart_proba(corpus,taille_corpus,my_format):NULL);
 }

/* on alloue le bilan provisoire */
bilan_tmp=(type_succes*)my_malloc(sizeof(type_succes)*taille_corpus);

/* On pose les questions possibles */

for(n=0;n<NbQuestion;n++)
 {
 if ((Debug>=2)&&((n%1000)==0))
  printf("\tquestion en cours : question=%d/%d (%.2f%%)\n",n,NbQuestion,(float)(n*100)/(float)NbQuestion);

 Question[n].nb=0;
 Question[n].nb_devt=0;
 Question[n].critere_oui=Question[n].critere_pere_oui=MaxCritere;
 Question[n].critere_non=Question[n].critere_pere_non=MaxCritere;

 if (Debug>=4) { printf("%d] ",n+1); AfficheQuestion(Question[n],my_format); }

 /* on pose la meme question sur tout le corpus */
 for (pt=corpus,nb_corpus=0;pt;pt=pt->suiv,nb_corpus++)
  {
  /* je stocke l'adresse de l'exemple dans le tableau local */
  CorpusTrainingLocal[nb_corpus]=pt;

  if (Debug>=4) { printf("%d] ",n+1); AfficheQuestion(Question[n],my_format); affiche_item(pt->item,my_format); }

  resu=PoseQuestion(pt,n,my_format);
   
  if (Debug>=4) fprintf(stdout,"  %s\n",resu!=NULL?"OUI":"NON");

  bilan_tmp[nb_corpus].good=resu!=NULL?True:False;
  bilan_tmp[nb_corpus].ptKS=resu;
  Question[n].nb+=(resu!=NULL?1:0);
  }

 if (nb_corpus!=taille_corpus) ERREUR("Prob dans les taille de corpus","");

 /* si le question permet d'obtenir 2 sous-corpus de taille > nb_elem_feuille, on la garde */

 if ((Question[n].nb>=nb_elem_feuille)&&((taille_corpus-Question[n].nb)>=nb_elem_feuille))
  {
  if (Debug>=2)
   {
   printf("Cette question separe suffisemment le corpus -> nb=%d taille_corpus=%d elem_feuille=%d\n",
    Question[n].nb,taille_corpus,nb_elem_feuille);
   AfficheQuestion(Question[n],my_format);
   }
  
  /* on alloue les bilans de la question et on recopie le bilan provisoire */
  Question[n].bilan=(type_succes*)my_malloc(sizeof(type_succes)*taille_corpus);
  if (CorpusDevt) Question[n].bilan_devt=(type_succes*)my_malloc(sizeof(type_succes)*taille_corpus_devt);
  else Question[n].bilan_devt=NULL;

  for (i=0;i<taille_corpus;i++) Question[n].bilan[i]=bilan_tmp[i];
  
  if (Debug>=2) printf("on repose la question sur tout le corpus de devt (%d elements)\n",taille_corpus_devt);
  /* on pose la meme question pour obtenir le corpus de devt (si le corpus de devt existe !!) */

  /*if (!strcmp(renvoi_chaine(Question[n].q,Question[n].niveau_q,my_format),"Are")) Debug=4; else Debug=2;*/

  if (CorpusDevt)
   {
   for (nb_corpus=0,pt=corpus_devt;pt;pt=pt->suiv,nb_corpus++)
    {
    CorpusDevtLocal[nb_corpus]=pt;

    if (Debug>2) affiche_item(pt->item,my_format);

    resu=PoseQuestion(pt,n,my_format);

    if (Debug>2) printf("\treponse = %s nb_corpus=%d\n",resu?"OUI":"NON",nb_corpus);
   
    Question[n].bilan_devt[nb_corpus].good=resu!=NULL?True:False;
    Question[n].bilan_devt[nb_corpus].ptKS=resu;
    Question[n].nb_devt+=(resu!=NULL?1:0);
    }
   }


   if ((CorpusDevt)&&((Question[n].nb_devt==0)||((taille_corpus_devt-Question[n].nb_devt)==0)))
    {
    if (Debug>=2) { printf("\t-> le corpus devt =0 , on abandonne la question\n"); }
    Question[n].critere_oui=Question[n].critere_non=MaxCritere;
    }
   else
    calcule_critere(n,taille_corpus,taille_corpus_devt,my_format);

  if (Debug>=2)
   {
   AfficheQuestion(Question[n],my_format);
   printf("BILAN : nb_oui=%d nb_oui_dev=%d pp=%.2lf  -  nb_non=%d nb_non_dev=%d pp=%.2lf\n",
    Question[n].nb,Question[n].nb_devt,Question[n].critere_oui,
    taille_corpus-Question[n].nb,taille_corpus_devt-Question[n].nb_devt,Question[n].critere_non);
   }
  }
 else
  {
  if (Debug>=2) printf("\t\t=> on met a zero Question[%d]=0\n",n);

  Question[n].nb=0;
  Question[n].bilan=Question[n].bilan_devt=NULL;
  }
 }

/* on libere le bilan provisoire */
free(bilan_tmp);

/* On choisit l'une des questions */
nchoix=choisit_question(critere_pere,taille_corpus,taille_corpus_devt,&best,my_format);

if (Debug>=2)
 {
 printf("on a choisi la question %d avec nb train=%d - best=%.2lf\n",nchoix,Question[nchoix].nb,best);
 printf("\t et nb devt = %d et taille_devt = %d\n",Question[nchoix].nb_devt,taille_corpus_devt);
 AfficheQuestion(Question[nchoix],my_format);
 }

/* si on a pas trouve de question qui separe le corpus ou diminue la pp alors on arrete : on fait une feuille */
if ((best==MaxCritere) || /* pas de reduction */
    (Question[nchoix].nb<nb_elem_feuille)||((taille_corpus-Question[nchoix].nb)<=nb_elem_feuille) || /* corpus train trop petit */
	((CorpusDevt)&&(((taille_corpus_devt-Question[nchoix].nb_devt)==0)||(Question[nchoix].nb_devt==0)))) /* corpus devt reduit a 0 */
 {
 if (Debug) printf("FEUILLE\t%s\n",best==MaxCritere?"best==MaxCritere":"taille<nb_elem_feuille");
 return nouveau_node(t_question_prec,'L',NUMBER_NODE++,'0','0',0,'0', /* 0 correspond au mot inconnu */
	my_format->mode==SUPERVISE?calcule_repart_proba(corpus,taille_corpus,my_format):NULL);
 }

if (Debug)
 {
 if (my_format->mode==NON_SUPERVISE)
  printf("NOEUD\tML_pere(oui)=%.2lf ML_oui(oui)=%.2lf - ML_pere(non)=%.2lf ML_non(non)=%.2lf - gain=%.2lf\n",
   Question[nchoix].critere_pere_oui,Question[nchoix].critere_oui,
   Question[nchoix].critere_pere_non,Question[nchoix].critere_non,
   (Question[nchoix].critere_pere_oui-Question[nchoix].critere_oui)+
   (Question[nchoix].critere_pere_non-Question[nchoix].critere_non));
 else
  printf("NOEUD\treference=%.2lf - oui=%.2lf - non=%.2lf\n",critere_pere,
	Question[nchoix].critere_oui,Question[nchoix].critere_non);
 fprintf(stdout,"\tquestion choisie (%d) :    \t",nchoix);
 AfficheQuestion(Question[nchoix],my_format);
 }

/* On cree la racine */
racine=nouveau_node(t_question_prec,'I',NUMBER_NODE++,Question[nchoix].t_question,Question[nchoix].qloc,
		Question[nchoix].q,Question[nchoix].niveau_q,NULL);


/*  SUR L'APPRENTISSAGE  */

/* On mets a jour les 2 listes d'exemples */
if (Debug>=2) printf("Mise a jour des listes good et baad\n");

for(i=nb_oui=nb_non=0,good=ptgood=baad=ptbaad=NULL;i<taille_corpus;i++)
 if (Question[nchoix].bilan[i].good==True)
  /* list GOOD */
  {
  nb_oui++;
  pt=CorpusTrainingLocal[i];
  if (Question[nchoix].t_question!='G')		/* not a global question, we update the KS */
   {
   Question[nchoix].bilan[i].ptKS->inKS=1;
   pt->KS.gap_avant=Question[nchoix].gap_avant;
   pt->KS.gap_apres=Question[nchoix].gap_apres;
   pt->KS.ptKS=Question[nchoix].bilan[i].ptKS;
   }
  if (good==NULL) good=ptgood=pt;
  else
   { ptgood->suiv=pt; ptgood=ptgood->suiv; }
  }
 else
  /* list BAAD */
  {
  nb_non++;
  pt=CorpusTrainingLocal[i];
  if (baad==NULL) baad=ptbaad=pt;
  else 
   { ptbaad->suiv=pt; ptbaad=ptbaad->suiv; }
  }
ptgood->suiv=ptbaad->suiv=NULL;

if (Debug) printf("\tcorpus train : OUI=%d NON=%d\n",nb_oui,nb_non);

/*  SUR LE DEVELOPPEMENT  */

if (CorpusDevt)
 {
 for(i=nb_oui_devt=nb_non_devt=0,devt_good=ptgood=devt_baad=ptbaad=NULL;i<taille_corpus_devt;i++)
  if (Question[nchoix].bilan_devt[i].good==True)
   /* list GOOD */
   {
   nb_oui_devt++;
   pt=CorpusDevtLocal[i];
   if (Question[nchoix].t_question!='G')         /* not a global question, we update the KS */
    {
    Question[nchoix].bilan_devt[i].ptKS->inKS=1;
    pt->KS.gap_avant=Question[nchoix].gap_avant;
    pt->KS.gap_apres=Question[nchoix].gap_apres;
    pt->KS.ptKS=Question[nchoix].bilan_devt[i].ptKS;
    }
   if (devt_good==NULL) devt_good=ptgood=pt;
   else
    { ptgood->suiv=pt; ptgood=ptgood->suiv; }
   }
  else
   /* list BAAD */
   {
   nb_non_devt++;
   pt=CorpusDevtLocal[i];
   if (devt_baad==NULL) devt_baad=ptbaad=pt;
   else 
    { ptbaad->suiv=pt; ptbaad=ptbaad->suiv; }
   }
 ptgood->suiv=ptbaad->suiv=NULL;

 if (Debug) printf("\tcorpus devt : OUI=%d NON=%d\n",nb_oui_devt,nb_non_devt);
 }
else
 { devt_good=devt_baad=NULL; nb_oui_devt=nb_non_devt=0; }

if (Debug==2)
 {
 printf("Les exemples good sont : \n");
 AfficheListSample(good,my_format);
 printf("Les exemples baad sont : \n");
 AfficheListSample(baad,my_format);
 }

/* sauvegarde rapport au tableau global 'Question' */
critere_oui=Question[nchoix].critere_oui;
critere_non=Question[nchoix].critere_non;

/* on libere les bilans */
for(i=0;i<NbQuestion;i++)
 {
 if (Question[i].bilan) free(Question[i].bilan);
 if (Question[i].bilan_devt) free(Question[i].bilan_devt);
 }

racine->fg=CreationTree(Question[nchoix].t_question,good,nb_oui,devt_good,nb_oui_devt,
			hauteur+1,nb_elem_feuille,haut_maxi,critere_oui,my_format);
racine->fd=CreationTree(t_question_prec,baad,nb_non,devt_baad,nb_non_devt,hauteur+1,nb_elem_feuille,haut_maxi,critere_non,my_format);

return racine;
}

/* pose une question sur un corpus, mets a jour les ptKS des exemples OUI et renvoi good et baad */

void separe_corpus(int n,ty_sample corpus,ty_sample *good,int *nbgood,ty_sample *baad,int *nbbaad, ty_format my_format)
{
ty_sample pt,ptgood,ptbaad;
ty_item resu;

/* on pose la question sur le corpus */
*good=ptgood=*baad=ptbaad=NULL;
*nbgood=*nbbaad=0;
for (pt=corpus;pt;pt=pt->suiv)
 {
 resu=PoseQuestion(pt,n,my_format);
 if (resu)
  {
  (*nbgood)++;
  if (Question[n].t_question!='G')	/* if not global, we modify the ptKS */
   {
   pt->KS.gap_avant=Question[n].gap_avant;
   pt->KS.gap_apres=Question[n].gap_apres;
   pt->KS.ptKS=resu;
   pt->KS.ptKS->inKS=1;
   }
  if (*good==NULL) *good=ptgood=pt;
  else { ptgood->suiv=pt; ptgood=ptgood->suiv; }
  }
 else
  {
  (*nbbaad)++;
  if (*baad==NULL) *baad=ptbaad=pt;
  else { ptbaad->suiv=pt; ptbaad=ptbaad->suiv; }
  }
 }
if (ptgood) ptgood->suiv=NULL;
if (ptbaad) ptbaad->suiv=NULL;
}

/*................................................................*/

/* creation d'un arbre a partir d'un fichier seed */

ty_node initialise_arbre(FILE *file,char t_quest_pere,int nb_elem_feuille,int haut_maxi,
	ty_sample corpus,           /* corpus d'apprentissage */
	int taille_corpus,          /* taille du corpus a l'entree ds la fonction */
	ty_sample corpus_devt,      /* corpus de developpement, ou NULL */
	int taille_corpus_devt,     /* taille du corpus developpement a l'entree ds la fonction */
	ty_format my_format)	    /* format de description du SCT */
{
char ch[TailleLigne],t_noeud,t_question,qloc,q[TailleLigne],old_tq,niveau_q;
ty_node pt;
wrd_index_t code_q;
int i,n,int_niveau_q,cate;

if (fgets(ch,TailleLigne,file))
 {
 i=sscanf(ch,"%c %d %c %s %d %c %c",&t_noeud,&cate,&old_tq,q,&int_niveau_q,&t_question,&qloc);
 if (i!=7)
  {
  sprintf(q,"%d elements trouves dans : %s (au lieu de 7)",i,ch);
  ERREUR("Mauvais format dans le fichier arbre",q);
  }

 niveau_q=(char)int_niveau_q-1;

 if (!strcmp(q,"NULL")) code_q=0; /* on mets le <UNK> sur NULL */
 else
  if (!Mot2Code(q,&code_q,my_format->tablex[niveau_q])) ERREUR("Item inconnu : ",q);

 if (t_noeud=='L') /* on est sur une feuille : on lance la creation de l'arbre */
  {
  double critere_pere;

  if (my_format->mode==NON_SUPERVISE)
   /*critere_pere=corpus_devt?calcule_perplexity_test(RacineOuiGram,corpus_devt,0,my_format):calcule_perplexity(RacineOuiGram);*/ 

   critere_pere=corpus_devt?calcule_perplexity_test(RacineOuiGram,corpus_devt,0,my_format):
		calcule_perplexity_test(RacineOuiGram,corpus,0,my_format);
  else
   critere_pere=calcule_gini_corpus(corpus_devt?corpus_devt:corpus,corpus_devt?taille_corpus_devt:taille_corpus,my_format);
  return CreationTree(t_quest_pere,corpus,taille_corpus,corpus_devt,taille_corpus_devt,0,nb_elem_feuille,haut_maxi,critere_pere,my_format);
  }
 else
  {
  int nbgood,nbbaad,dev_nbgood,dev_nbbaad;
  ty_sample good,baad,dev_good,dev_baad;

  pt=nouveau_node(t_quest_pere,t_noeud,cate,t_question,qloc,code_q,niveau_q,NULL);

  /* on recherche la question dans la liste des questions */
  for(n=0;(n<NbQuestion)&&
	((Question[n].q!=code_q) ||
	 (Question[n].niveau_q!=(int)(niveau_q)) ||
	 (Question[n].t_question!=t_question) ||
	 (Question[n].qloc!=qloc));n++);
  if (n==NbQuestion) ERREUR("je ne trouve pas la question :",q);

  /* on pose la question sur le corpus d'apprentissage */
  separe_corpus(n,corpus,&good,&nbgood,&baad,&nbbaad,my_format);

  if (Debug)
   {
   printf("Question SEED :\n");
   AfficheQuestion(Question[n],my_format);
   printf("Nb good=%d  -  Nb baad=%d\n",nbgood,nbbaad);
   }

  /* on pose la question sur le corpus de developpement s'il existe */
  if (corpus_devt) separe_corpus(n,corpus_devt,&dev_good,&dev_nbgood,&dev_baad,&dev_nbbaad,my_format);
  else { dev_good=dev_baad=NULL; dev_nbgood=dev_nbbaad=0; }

  pt->fg=initialise_arbre(file,t_question,nb_elem_feuille,haut_maxi,good,nbgood,dev_good,dev_nbgood,my_format);
  pt->fd=initialise_arbre(file,t_quest_pere,nb_elem_feuille,haut_maxi,baad,nbbaad,dev_baad,dev_nbbaad,my_format);
  return pt;
  }
 }
ERREUR("ON NE DEVRAIT JAMAIS ETRE LA","");

return NULL;
}

ty_node LanceCreationTree
	(
	int nb_elem_feuille,
	int haut_maxi,
	char *seedfile,
	ty_format my_format
	)
{
ty_node racine;
double critere_pere;

MaxHauteur=0;

if (my_format->mode==SUPERVISE)
 {
 RepartEtiquetteOUI=(double*)malloc(sizeof(double)*my_format->nb_etiquette);
 RepartEtiquetteNON=(double*)malloc(sizeof(double)*my_format->nb_etiquette);
 }

/* si un fichier arbre SEED est defini, on le charge en mettant a jour les corpus */

if (strcmp(seedfile,"NULL"))
 {
 FILE *file;
 if (Debug) fprintf(stderr,"Lecture du fichier SEED : %s\n",seedfile);

 if (!(file=fopen(seedfile,"rt"))) ERREUR("Can't read the seed tree file :",seedfile);
 racine=initialise_arbre(file,'X',nb_elem_feuille,haut_maxi,
		  CorpusTraining[0],NbSampleTraining,CorpusDevt?CorpusDevt[0]:NULL,NbSampleDevt,my_format);
 fclose(file);
 }
else
 {
 if (my_format->mode==NON_SUPERVISE)
  critere_pere=CorpusDevt?calcule_perplexity_test(RacineOuiGram,CorpusDevt[0],0,my_format):calcule_perplexity(RacineOuiGram);
 else
  critere_pere=calcule_gini_corpus(CorpusDevt?CorpusDevt[0]:CorpusTraining[0],CorpusDevt?NbSampleDevt:NbSampleTraining,my_format);

 if (Debug) printf("Critere initial = %.2lf\n",critere_pere);

 racine=CreationTree('X',CorpusTraining[0],NbSampleTraining,
		    CorpusDevt?CorpusDevt[0]:NULL,NbSampleDevt,
		    0,nb_elem_feuille,haut_maxi,
		    critere_pere,my_format);
 }
if (Debug) fprintf(stderr,"\nCreation terminee. Hauteur maximale de l'arbre : %d\n",MaxHauteur);
return racine;
}

/*................................................................*/

ty_format lis_descripteur(char *fich_desc,int *nb_elem_feuille,char *fich_arbre,char *seed_file)
{
FILE *file;
char ch[TailleLigne],*pt,ch2[TailleLigne];
int si_arbre=0,si_corpus_train=0,si_corpus_dev=0,si_taille_feuille=0;
ty_format my_format;

/* le champs SEED n'est pas obligatoire */
strcpy(seed_file,"NULL");

if (!(file=fopen(fich_desc,"rt"))) ERREUR("can't open :",fich_desc);
while(fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"ARBRE",5))
  {
  si_arbre=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(fich_arbre,pt);
  if (Debug) fprintf(stderr,"\t* Lecture descripteur\n");
  sprintf(ch2,"%s.desc_lex",pt);
  my_format=LectureDescripteurFormat(ch2);

  if (Debug) fprintf(stderr,"\t* Lecture de la liste de questions\n");
  LectureListQuestion(pt,my_format);
  } else
 if (!strncmp(ch,"CORPUS_TRAIN",12))
  {
  if (!si_arbre) ERREUR("'ARBRE' doit etre defini en premier dans ",fich_desc);
  si_corpus_train=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (Debug) fprintf(stderr,"\t* Lecture du fichier training");
  lecture_corpus_training(pt,my_format); if (Debug) fprintf(stderr," -> (%d samples read) OK\n",NbSampleTraining);
  } else
 if (!strncmp(ch,"CORPUS_DEV",10))
  {
  if (!si_arbre) ERREUR("'ARBRE' doit etre defini en premier dans ",fich_desc);
  si_corpus_dev=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (Debug) fprintf(stderr,"\t* Lecture du fichier developpement");
  lecture_corpus_devt(pt,my_format); if (Debug) fprintf(stderr," -> (%d samples read) OK\n",NbSampleDevt);
  } else
 if (!strncmp(ch,"TAILLE_FEUILLE",14))
  {
  if (!si_arbre) ERREUR("'ARBRE' doit etre defini en premier dans ",fich_desc);
  si_taille_feuille=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (sscanf(pt,"%d",nb_elem_feuille)!=1) ERREUR("dans 'TAILLE_FEUILLE' du descripteur:",fich_desc);
  } else
 if (!strncmp(ch,"SEED",4))
  {
  if (!si_arbre) ERREUR("'ARBRE' doit etre defini en premier dans ",fich_desc);
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(seed_file,pt);
  }
 }
if (!si_arbre) ERREUR("il manque ARBRE dans ",fich_desc);
if (!si_corpus_train) ERREUR("il manque CORPUS_TRAIN dans ",fich_desc);
if (!si_corpus_dev) ERREUR("il manque CORPUS_DEV dans ",fich_desc);
if (!si_taille_feuille) ERREUR("il manque TAILLE_FEUILLE dans ",fich_desc);

return my_format;
}

/*................................................................*/

int main(argc,argv)
 int argc;
 char **argv;
{
ty_node racine;
FILE *fileout;
int nb_elem_feuille,haut_maxi;
char fich_arbre[TailleLigne],seed_file[TailleLigne];
ty_format my_format;

if ((argc==1)||(!strcmp(argv[1],"-h")))
 {
 fprintf(stderr,"Syntaxe : %s [-h] <fich descripteur>\n",argv[0]);
 if ((argc>1)&&(!strcmp(argv[1],"-h")))
  {
  fprintf(stderr,"\t-> permet de construire un arbre SCT en fonction d'un fichier descripteur\n");
  }
 exit(0);
 }

if (Debug) fprintf(stderr,"Initialisation :\n");
my_format=lis_descripteur(argv[1],&nb_elem_feuille,fich_arbre,seed_file);
if (Debug) fprintf(stderr,"Terminee\n\n");

haut_maxi=10000; /* on ne controle pas la hauteur */

/* creation de l'arbre et sauvegarde */

if (Debug) fprintf(stderr,"Creation de l'arbre :\n");

if (!(fileout=fopen(fich_arbre,"w"))) ERREUR("Je ne peux ecrire dans",fich_arbre);
racine=LanceCreationTree(nb_elem_feuille,haut_maxi,seed_file,my_format);
AfficheArbre(racine,fileout,my_format);
fclose(fileout);
if (Debug) fprintf(stderr,"C'est fini\n");
}
 
