/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  Construit les Modeles de Langage ARPA a partir des corpus feuilles
    generes par 'lia_sct_use'  */


/*  Utilise les arbres CART sur les phonemes construit avec np_build_cart  */
/*  Fred 0398 - Modif 0699  - Modif 0100  - Modif 0500  -  Modif 1200 (devt)  */

#include <lia_sct.h>

/* si on evalue le gain avant de descendre ds l'arbre en produisant les ARPA */
#define REEVALUE_PPLEX	0

/*................................................................*/

/* calcul des pp moyennes et des gains moyens */

double Total_refe_entropy,Total_test_entropy;
int Nb_2gram_test;

/*................................................................*/

/*  Codage de la repartition des exemples apres le parcours de l'arbre  */

typedef struct
	{
	ty_node feuille;
	int nb;
	ty_sample sample;
	FILE *file;
	} type_groupe;

int NbGroupe;
type_groupe *TableGroupe;

/*................................................................*/

/*  Fonctions d'affichage  */

int AfficheFeuilleIndice(feuille,indice,nom_corpus,file_corpus)
 ty_node feuille;
 int indice;
 char *nom_corpus;
 FILE *file_corpus;
{
int n;
FILE *file;
static char ch[TailleLigne];

for(n=0;(n<NbGroupe)&&(TableGroupe[n].feuille!=feuille);n++);
if (n==NbGroupe) /*printf("Cette feuille n'a recu aucun exemple ....\n")*/;
else
 {
 /* on lis les fichiers correspondant */
 sprintf(ch,"%s/feuille%d.exp",nom_corpus,n);
 /*printf("ouverture de : %s\n",ch);*/

 if (file=fopen(ch,"rt"))
  {
  while (fgets(ch,TailleLigne,file))
   {
   indice++;
   fprintf(file_corpus,"%s",ch);
   }
  fclose(file);
  }
 }
return indice;
}

int AfficheCorpusSousArbre(racine,indice,nom_corpus,file_corpus)
 ty_node racine;
 int indice;
 char *nom_corpus;
 FILE *file_corpus;
{
if (racine->t_noeud=='L') return AfficheFeuilleIndice(racine,indice,nom_corpus,file_corpus);
else
 {
 indice=AfficheCorpusSousArbre(racine->fg,indice,nom_corpus,file_corpus);
 indice=AfficheCorpusSousArbre(racine->fd,indice,nom_corpus,file_corpus);
 return indice;
 }
}

void read_resu_pplex_CMU(char *nomfic,double *pplex,double *entropy,int *nb2gram)
{
FILE *file;
char ch[TailleLigne];

if (!(file=fopen(nomfic,"rt"))) ERREUR("Can't read:",nomfic);
while(fgets(ch,200,file))
 {
 if (!strncmp(ch,"Perplexity",10))
  if (sscanf(ch,"Perplexity = %lf, Entropy = %lf bits",pplex,entropy)!=2) ERREUR("bad pplex file:",nomfic);
 if (!strncmp(ch,"Computation",11))
  if (sscanf(ch,"Computation based on %d words.",nb2gram)!=1) ERREUR("bad pplex file:",nomfic);
 }
fclose(file);
}

void AfficheCorpusNiveau(
	int *num_noeud,
 	ty_node pt,
 	int min_nbex,
 	char *nom_corpus,
 	char *nom_noeud,
 	char *nom_devt,
 	int si_test,
 	double prev_pp,
	int on_continue,
 	char *ch_arbre,
	FILE *fich_feuille
	)
{
static char ch[200];
int nb_item_corpus,nouv_on_continue,nb_item_corpus_dev,nouv_nb2gram;
FILE *file;
double nouv_pp,refe_pp,nouv_entropy,refe_entropy;

if (pt)
 {
 (*num_noeud)++;

 if (si_test)
  {
  /* on sort le corpus de test correspondant aux noeuds de l'arbre */
  sprintf(ch,"%s/noeud_n%d.test",nom_noeud,(*num_noeud));
  if (!(file=fopen(ch,"wt"))) ERREUR("can't open :",ch);
  nb_item_corpus=AfficheCorpusSousArbre(pt,0,nom_corpus,file);
  fclose(file);

  if (nb_item_corpus)
   {
   /* on teste si le fichier arpa de ce corpus existe */
   sprintf(ch,"%s/noeud_n%d.arpa",nom_noeud,(*num_noeud));
   if (file=fopen(ch,"r"))
    {
    fclose(file);

    /* si le suivant existe, on descend sans rien faire sauf si on est sur une feuille */
    sprintf(ch,"%s/noeud_n%d.arpa",nom_noeud,(*num_noeud)+1);
    if ((pt->t_noeud=='L')||(!(file=fopen(ch,"r"))))
     {
     /* on affiche les renseignements */
     printf("Corpus %d : taille(test)=%d\n",(*num_noeud),nb_item_corpus);

     /* on ecris dans fich_feuille si le fichier est ouvert */
     if (fich_feuille) fprintf(fich_feuille,"%d\n",(*num_noeud));

     /* on calcule la pplex du fichier test avec le modele noeud */
     sprintf(ch,"csh $LIA_SCT_REP/script/lance_cmu_test %s.desc_lex %s/noeud_n%d %s/noeud_n0.arpa > /dev/null",
      ch_arbre,nom_noeud,(*num_noeud),nom_noeud);
     system(ch);

     /* on lit les resultats de la pplex */
     
     sprintf(ch,"%s/noeud_n%d.noeud.pplex",nom_noeud,(*num_noeud));
     read_resu_pplex_CMU(ch,&nouv_pp,&nouv_entropy,&nouv_nb2gram);
     sprintf(ch,"%s/noeud_n%d.global.pplex",nom_noeud,(*num_noeud));
     read_resu_pplex_CMU(ch,&refe_pp,&refe_entropy,&nouv_nb2gram);
     
     /* on nettoie */
     sprintf(ch,"rm %s/noeud_n%d.noeud.pplex",nom_noeud,(*num_noeud));
     system(ch);
     sprintf(ch,"rm %s/noeud_n%d.global.pplex",nom_noeud,(*num_noeud));
     system(ch);

     /* affichage des resultats */
     printf("\tPerplexite : modele noeud = %.2lf  -  modele reference = %.2lf  -  gain = %.2lf%%\n",
	nouv_pp,refe_pp,((refe_pp-nouv_pp)*(double)100)/refe_pp);

     /* on mets a jour les compteurs generaux */
     Total_refe_entropy+=refe_entropy*nouv_nb2gram;
     Total_test_entropy+=nouv_entropy*nouv_nb2gram;
     Nb_2gram_test+=nouv_nb2gram;
     }
    else
     {
     /* si on n'est pas sur une feuille on qu'on a reussit a ouvrir le fichier arpa suivant */
     fclose(file);
     }
    }
   else
    {
    fprintf(stderr,"Warning : le fichier %s n'existe pas\n",ch);
    if (Debug>=2)
      {
      printf("Warning : le fichier %s n'existe pas\n",ch);
      printf("on efface le fichier %s/noeud_n%d.test et %s/noeud_n%d.corpus\n",nom_noeud,(*num_noeud),nom_noeud,(*num_noeud));
      }
    sprintf(ch,"rm %s/noeud_n%d.test",nom_noeud,(*num_noeud)); system(ch);
    /* on efface aussi le corpus */ 
    sprintf(ch,"rm %s/noeud_n%d.corpus",nom_noeud,(*num_noeud)); system(ch);
    }
   }
  else
   {
   if (Debug>=2)
    printf("TEST : il n'y a pas d'item de test pour le corpus n%d\n",(*num_noeud));
   /* on efface le fichier test qui est vide */
   sprintf(ch,"rm %s/noeud_n%d.test",nom_noeud,(*num_noeud)); system(ch);
   /* on efface les autres ? */
   /*
   sprintf(ch,"rm %s/noeud_n%d.arpa",nom_noeud,(*num_noeud)); system(ch);
   sprintf(ch,"rm %s/noeud_n%d.corpus",nom_noeud,(*num_noeud)); system(ch);
   */
   }
  }
 else
  {
  /* on calcule un modele ARPA a partir du corpus */

  /* on cree les .corpus */
  sprintf(ch,"%s/noeud_n%d.corpus",nom_noeud,(*num_noeud));
  if (!(file=fopen(ch,"wt"))) ERREUR("can't open :",ch);
  nb_item_corpus=AfficheCorpusSousArbre(pt,0,nom_corpus,file);
  fclose(file);

  /* on cree les .dev si le rep est diff de NULL */
  if (strcmp(nom_devt,"NULL"))
   {
   sprintf(ch,"%s/noeud_n%d.dev",nom_noeud,(*num_noeud));
   if (!(file=fopen(ch,"wt"))) ERREUR("can't open :",ch);
   nb_item_corpus_dev=AfficheCorpusSousArbre(pt,0,nom_devt,file);
   fclose(file);
   }

  /* on affiche les renseignements */
  printf("Corpus %d : taille(train)=%d\n",(*num_noeud),nb_item_corpus);

  if ((nb_item_corpus>min_nbex)&&(on_continue))
   {
   nouv_on_continue=1;

   /* on calcule le ML ARPA */
   sprintf(ch,"csh $LIA_SCT_REP/script/lance_cmu %s.desc_lex %s/noeud_n%d > ./sasa_tmp",ch_arbre,nom_noeud,(*num_noeud));
   system(ch);
   system("rm ./sasa_tmp");

   if (REEVALUE_PPLEX)
    {
    /* on evalue la pplex du ML sur le corpus d'entrainement */
    sprintf(ch,"csh $LIA_SCT_REP/script/lance_cmu_test_training %s.desc_lex %s/noeud_n%d %s/noeud_n0.arpa > %s/noeud_n%d.soso_tmp",
      ch_arbre,nom_noeud,(*num_noeud),nom_noeud,nom_noeud,(*num_noeud));
    system(ch);
    sprintf(ch,"%s/noeud_n%d.soso_tmp",nom_noeud,(*num_noeud));
    if (!(file=fopen(ch,"rt"))) ERREUR("Can't read:",ch);
    while(fgets(ch,200,file)&&(!strstr(ch,"Perplexity")));
    if (sscanf(ch,"Perplexity = %lf",&nouv_pp)!=1) ERREUR("je ne peux lire la pplex ML noeud dans soso.tmp","");
    while(fgets(ch,200,file)&&(!strstr(ch,"Perplexity")));
    if (sscanf(ch,"Perplexity = %lf",&refe_pp)!=1) ERREUR("je ne peux lire la pplex ML ref dans soso.tmp","");
    fclose(file);
    sprintf(ch,"rm %s/noeud_n%d.soso_tmp",nom_noeud,(*num_noeud));
    system(ch);

    /* affichage des resultats en perplexite */
    printf("\tpplex(pere)=%.2lf - pplex(ML%d)=%.2lf - pplex(ML0)=%.2lf\n",prev_pp,(*num_noeud),nouv_pp,refe_pp);

    /* Si on perds, on efface les fichiers arpa !! */
    if ((nouv_pp>prev_pp)||(nouv_pp>refe_pp))
     {
     printf("on y perd : on ne descends plus dans l'arbre\n");
     sprintf(ch,"rm %s/noeud_n%d.arpa",nom_noeud,(*num_noeud)); system(ch);
     sprintf(ch,"rm %s/noeud_n%d.corpus",nom_noeud,(*num_noeud)); system(ch);
     nouv_on_continue=0;
     }
    else printf("on y gagne : on continue a descendre dans l'arbre\n");
    }
   }
  else
   {
   nouv_on_continue=0;

   if (nb_item_corpus<=min_nbex)
    {
    sprintf(ch,"rm %s/noeud_n%d.corpus",nom_noeud,(*num_noeud)); system(ch);
    printf("nb d'item d'apprentissage insufisante : on ne descends plus dans l'arbre\n");
    }
   else
    printf("on ne fait rien, car les niveaux superieurs l'on decide ainsi\n");
   }
  }
 if (pt->t_noeud=='I')
  {
  AfficheCorpusNiveau(num_noeud,pt->fg,min_nbex,nom_corpus,nom_noeud,nom_devt,
	si_test,nouv_pp,nouv_on_continue,ch_arbre,fich_feuille);

  AfficheCorpusNiveau(num_noeud,pt->fd,min_nbex,nom_corpus,nom_noeud,nom_devt,
	si_test,prev_pp,on_continue,ch_arbre,fich_feuille);
  }
 }
}

/*................................................................*/

/*  Lecture d'un fichier CART et construction de l'arbre  */

ty_node ConstruitArbre(FILE *file,char KS,ty_format my_format)
{
char ch[TailleLigne],t_noeud,t_question,qloc,q[TailleLigne],old_KS;
ty_node pt;
wrd_index_t code_q;
int i,int_niveau_q,cate;

if (fgets(ch,TailleLigne,file))
 {
 i=sscanf(ch,"%c %d %c %s %d %c %c",&t_noeud,&cate,&old_KS,q,&int_niveau_q,&t_question,&qloc);
 if (i!=7)
  {
  sprintf(q,"%d elements trouves dans : %s (au lieu de 7)",i,ch);
  ERREUR("Mauvais format dans le fichier arbre",q);
  }

 int_niveau_q=(char)int_niveau_q-1;

 if (!strcmp(q,"NULL")) code_q=0; /* on mets le <UNK> sur NULL */
 else
  if (!Mot2Code(q,&code_q,my_format->tablex[int_niveau_q])) ERREUR("Item inconnu : ",q);
 pt=nouveau_node(KS,t_noeud,cate,t_question,qloc,code_q,int_niveau_q,NULL/* a remplacer en mode SUPERVISE */);
 if (t_noeud=='I')
  {
  pt->fg=ConstruitArbre(file,t_question,my_format);
  pt->fd=ConstruitArbre(file,KS,my_format);
  }
 else NbGroupe++;
 return pt;
 }
return NULL;
}

void AffecteFeuille(racine,i)
 ty_node racine;
 int *i;
{
if (racine)
 {
 if (racine->t_noeud=='L') /* c'est une feuille */
  {
  if (*i==NbGroupe) { fprintf(stderr,"NbGroupe trop petit !!\n"); exit(0); }
  TableGroupe[*i].feuille=racine;
  TableGroupe[*i].nb=0;
  TableGroupe[*i].sample=NULL;
  TableGroupe[*i].file=NULL;
  (*i)++;
  }
 AffecteFeuille(racine->fg,i);
 AffecteFeuille(racine->fd,i);
 }
}

ty_node LectureTree(char *chfile,ty_format my_format)
{
FILE *file;
int n;
ty_node racine;

if (!(file=fopen(chfile,"rt"))) ERREUR("Can't open : ",chfile);

/*  La KS a la racine est X cad <+>  */
NbGroupe=0;
racine=ConstruitArbre(file,'X',my_format);
fclose(file);

/* on alloue le tableau des groupes d'item */
TableGroupe=(type_groupe*)malloc(sizeof(type_groupe)*NbGroupe);
n=0;
AffecteFeuille(racine,&n);
return racine;
}

/*................................................................*/

void lis_descripteur(char *fich_desc,ty_node *racine,
		     char *nom_arbre,char *nom_noeud,char *nom_devt,
		     char *nom_feuille_train,char *nom_feuille_test)
{
FILE *file;
char ch[TailleLigne],*pt,ch2[TailleLigne];
int si_arbre=0,si_noeud=0,si_devt=0,si_cp_dev=0;
ty_format my_format;

if (!(file=fopen(fich_desc,"rt"))) { fprintf(stderr,"can't open : %s\n",fich_desc); exit(0); }
while(fgets(ch,TailleLigne,file))
 {
 if (!strncmp(ch,"ARBRE",5))
  {
  si_arbre=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  sprintf(ch2,"%s.desc_lex",pt);
  my_format=LectureDescripteurFormat(ch2);
  sprintf(ch2,"%s.ref",pt);
  my_format->tablex[my_format->nb_niveaux]=ChargeLexique(ch2);
  (*racine)=LectureTree(pt,my_format);
  strcpy(nom_arbre,pt);
  } else
 if (!strncmp(ch,"NOEUD_REP",9))
  {
  si_noeud=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(nom_noeud,pt);
  } else
 if (!strncmp(ch,"DEVT_REP",8))
  {
  si_devt=1;
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(nom_devt,pt);
  } else
 if (!strncmp(ch,"CORPUS_DEV",10))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  if (strcmp(pt,"NULL")) si_cp_dev=1;
  } else
 if (!strncmp(ch,"FEUILLE_REP",11))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(nom_feuille_train,pt);
  } else
 if (!strncmp(ch,"TEST_REP",8))
  {
  pt=strtok(ch," \t\n"); pt=strtok(NULL," \t\n");
  strcpy(nom_feuille_test,pt);
  }
 }
if (!si_arbre) { fprintf(stderr,"ERREUR : il manque 'ARBRE' dans %s\n",fich_desc); exit(0); }
if (!si_noeud) { fprintf(stderr,"ERREUR : il manque 'NOEUD_REP' dans %s\n",fich_desc); exit(0); }
if (!si_devt)  { fprintf(stderr,"ERREUR : il manque 'DEVT_REP' dans %s\n",fich_desc); exit(0); }

/* si le corpus DEV est a NULL, on ne veut pas creer les fichiers .dev */
if (si_cp_dev==0) strcpy(nom_devt,"NULL");

fclose(file);
}

void main(argc,argv)
 int argc;
 char **argv;
{
ty_node racine;
char ch[TailleLigne],
     nom_arbre[TailleLigne],nom_noeud[TailleLigne],nom_devt[TailleLigne],
     nom_feuille_train[TailleLigne],nom_feuille_test[TailleLigne];
int n,min_nbex=0,num_noeud,just_test=False,just_arpa=False;
FILE *fich_feuille=NULL;
double pplex_refe,pplex_test;

if ((argc==1)||(!strcmp(argv[1],"-h")))
 {
 fprintf(stderr,"Syntaxe : %s [-h] -desc <fichier descripteur> [-min_exp nb] [-just_test] [-just_arpa]\n",argv[0]);
 exit(0);
 }

fprintf(stderr,"Chargement des modeles -> ");

for (n=1;n<argc;n++)
 if (!strcmp(argv[n],"-desc"))
  lis_descripteur(argv[++n],&racine,nom_arbre,nom_noeud,nom_devt,nom_feuille_train,nom_feuille_test);
 else
 if (!strcmp(argv[n],"-min_exp")) sscanf(argv[++n],"%d",&min_nbex);
 else
 if (!strcmp(argv[n],"-just_test")) just_test=True;
 else
 if (!strcmp(argv[n],"-just_arpa")) just_arpa=True;
 else
  {
  fprintf(stderr,"ERREUR : je ne connais pas l'option '%s' ....\n",argv[n]);
  exit(0);
  }

fprintf(stderr,"Termine\n");

Total_refe_entropy=Total_test_entropy=(double)0;
Nb_2gram_test=0;

/* le fichier .list_feuille precise la liste des ML que l'on a utilise pour
   calculer la pplex sur le test.
   Precisons que si cette liste n'est pas exhaustive, c'est qu'on a selectionne
   les feuilles par rapport aux transcriptions exactes du test. Il ne faut donc
   pas l'utiliser dans la reco. */
sprintf(ch,"%s.list_feuille",nom_arbre);
if (!(fich_feuille=fopen(ch,"wt"))) fich_feuille=NULL;

/* on fait d'abord les fichiers ARPA */
if (just_test==False)
 {
 num_noeud=-1;
 AfficheCorpusNiveau(&num_noeud,racine,min_nbex,nom_feuille_train,nom_noeud,nom_devt,
	0,(double)1000.0,1,nom_arbre,fich_feuille);
 }

/* puis l'extraction des fichiers tests */
if (just_arpa==False)
 {
 num_noeud=-1;
 AfficheCorpusNiveau(&num_noeud,racine,min_nbex,nom_feuille_test,nom_noeud,nom_devt,
	1,(double)1000.0,1,nom_arbre,fich_feuille);
 if (Nb_2gram_test)
  {
  pplex_refe=pow(2,Total_refe_entropy/(double)Nb_2gram_test);
  pplex_test=pow(2,Total_test_entropy/(double)Nb_2gram_test);
  printf("\n\n\n\nBILAN :\n");

  printf("Perplexite moyenne : modele noeud = %.2lf  -  modele reference = %.2lf  -  Gain = %.2lf%%\n",
   pplex_test,pplex_refe,((pplex_refe-pplex_test)*(double)100)/pplex_refe);

  }
 }

if (fich_feuille) fclose(fich_feuille);
}
  
