/*
     ------------------------------------------------
     LIA_SCT: A DECISION-TREE BASED STRING CLASSIFIER
     ------------------------------------------------
 
     Copyright (C) 2001 FREDERIC BECHET
 
     ..................................................................
 
     This file is part of LIA_SCT
 
     LIA_SCT is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.
 
     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.
 
     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
     ..................................................................
 
     For any publication related to scientific work using LIA_SCT,
     the following reference paper must be mentioned in the bibliography: 
         
     Bechet F., Nasr A., Genet F. 2000, "Tagging unknown proper names
     using decision trees", in 38th Annual Meeting of the Association for
     Computational Linguistics (ACL), Hong-Kong, China, 2000, pp. 77-84
     ..................................................................
                               
     Contact :
               FREDERIC BECHET - LIA - UNIVERSITE D'AVIGNON
               AGROPARC BP1228 84911  AVIGNON  CEDEX 09  FRANCE
               frederic.bechet@lia.univ-avignon.fr
     ..................................................................
*/
/*-----------------------------------------------*/
/*  LIA_SCT * Toolkit de gestion des arbres SCT  */
/*-----------------------------------------------*/
/* Version generique (V2.0) - Fevrier 2001 FRED  */
/*-----------------------------------------------*/

/*  lia_sct_pplex : gestion de la perplexite (ngram)  */

/*................................................................*/

/* gestion des ngram pour la perplexite - on prevoit jusqu'au 2grams */

#define MAX_PPLEX	((double)1000000)

#define CUTOFF_1	0
#define CUTOFF_2	0

#define PENAL_UNK	((double)1) /*10*/
#define PENAL_1		((double)2) /*5*/

typedef struct type_niveau
	{
	wrd_index_t gram;
	long nb;
	double pb;
	struct type_niveau *next,*fg,*fd;
	} *ty_niveau;

/*  On a 3 arbres globaux contenant les modeles ngrams :
    - le premier appele : RacineOuiGram correspondant aux item ayant repondu oui
    - le deuxieme       : RacineNonGram correspondant aux item ayant repondu non
    - le troisieme      : RacineAllGram groupant les 2 corpus pour evaluer la perplexite sur le noeud pere
    les 3 arbres ont une structure identique, c'est juste leur remplissage qui differe  */

extern ty_niveau RacineOuiGram,RacineNonGram,RacineAllGram;

/*................................................................*/

/* fonction de remplissage de l'arbre cree : compte et proba */

ty_niveau ajoute_item_ngram(ty_item,ty_niveau,long*,ty_format);
void rempli_proba(long,ty_niveau,int);
void raz_gram(ty_niveau); /* remet compte et proba a zero */

/*................................................................*/

/* Calcul des probas et de la perplexite */

double calcule_perplexity(ty_niveau);
double calcule_perplexity_test(ty_niveau,ty_sample,int,ty_format);
double calcule_perplexity_test_succes(char,ty_niveau,ty_sample *,type_succes *,int,int,ty_format);

/*................................................................*/

/* debugage */

void affiche_proba(char *,char *,char *,ty_niveau,ty_format);

/*................................................................*/

ty_niveau rempli_ngram(ty_item,ty_niveau,long *,ty_format);
 
/*................................................................*/

ty_niveau rempli_niveaux(ty_sample,ty_format);
 
